﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using System.IO;
using CCInfo = System.Tuple<string, int, long>;

namespace TilingsConsole
{
    class Program
    {
        static void Main(string[] args)
        {
            if (!Directory.Exists("Output"))
                Directory.CreateDirectory("Output");
            
            string filePath = "Output/output.txt";
            var boxDim = new BoxDimensions(2, 3, 4);
            var ccInfo = Region.ParallelFindFlipCC(boxDim);
            WriteCCsToFile(filePath, ccInfo);  
            
            //Tiling with a somewhat large component
            string tilingComponentZero = "RL,RL,RL;RL,RL,RL;DB,UD,BU;BF,BB,FB;FB,FF,BF;DF,UD,FU;RL,RL,RL;RL,RL,RL";
            VisitedCC(tilingComponentZero, ccInfo, "Output/outputComps.txt");

            // SVG Drawing of a tiling (2D representation)
            SVGDrawing.NewDrawTiling(tilingComponentZero, "Output/tilingExample.svg");
            Console.Read();

        }

        public static void FindComponentsAndWriteToFile(BoxDimensions boxDim, string filePath)
        {
            WriteCCsToFile(filePath, Region.ParallelFindFlipCC(boxDim));            
        }
        
        /// <summary>
        /// Write a summary of each connected component to filePath.
        /// Parameter componentsInfo is likely to have been obtained from another method, e.g. Region.ParallelFindFlipCC().
        /// </summary>
        /// <param name="filePath">The path of the output file</param>
        /// <param name="componentsInfo"></param>
        /// <param name="maxTolerance"></param>
        static void WriteCCsToFile(string filePath, LargeConnectedComponentsInfo componentsInfo, int maxTolerance = 1000000)
        {
            int i = 0;
            using (var writer = new StreamWriter(filePath, false))
            {

                foreach (StringPair sp in componentsInfo.KnownParents.Keys)
                {
                    Console.WriteLine(DateTime.Now);
                    writer.WriteLine(DateTime.Now);

                    Console.WriteLine("Considering CC with index {0}\r\nCC has {1} members\r\nParent = {2}\r\n", i, componentsInfo.KnownParents[sp], sp.ToTilingString());
                    
                    writer.WriteLine("----- Connected component {0} -----", i);
                    writer.WriteLine("Parent: {0}\r\nIndex: {1}\r\nNumber of Elems: {2}\r\n", sp.ToTilingString(), i, componentsInfo.KnownParents[sp]);
                    writer.WriteLine("Twist: {0}\r\n", Region.FromString(sp.ToTilingString()).Twist());
                    if (componentsInfo.KnownParents[sp] < maxTolerance)
                    {       
                         
                        HashSet<string> positives;
                        HashSet<string> negatives;

                        Region.FindTritsFromCC(sp.ToTilingString(), out positives, out negatives);

                        writer.WriteLine("Positive trits:");
                        HashSet<string> visitedParents = new HashSet<string>();

                        foreach (string til in positives)
                        {
                            var info = componentsInfo.GetConnectedComponent(til);

                            if (visitedParents.Add(info.Item1))
                                writer.WriteLine("Parent: {0}\r\nIndex: {1}\r\nNumber of Elems: {2}\r\n", info.Item1, info.Item2, info.Item3);
                        }

                        visitedParents.Clear();

                        writer.WriteLine("Negative trits:");
                        foreach (string til in negatives)
                        {
                            var info = componentsInfo.GetConnectedComponent(til);

                            if (visitedParents.Add(info.Item1))
                                writer.WriteLine("Parent: {0}\r\nIndex: {1}\r\nNumber of Elems: {2}\r\n", info.Item1, info.Item2, info.Item3);
                        }

                    }
                    else
                    {
                        writer.WriteLine("Number of elements exceeded maximum tolerance, trit calculation was skipped.\r\n");                    }

                    i++;
                }

                writer.Flush();
            }

            Console.WriteLine("Finished writing to file");
        }

        /// <summary>
        /// A helper method for treating large components (not too large, though). Attempts to find all the connected components  
        /// that are reached from the tiling parameters via flips and one trit, but may throw OutOfMemoryException if component is too large.
        /// Parameter componentsInfo is likely to have been obtained from another method, e.g. Region.ParallelFindFlipCC().
        /// </summary>
        /// <param name="tiling">The starting tiling</param>
        /// <param name="componentsInfo">The connected components info from the region (must be compatible with tiling)</param>
        /// <param name="filename">The file where to write the output.</param>
        static void VisitedCC(string tiling, LargeConnectedComponentsInfo componentsInfo, string filename)
        {

            HashSet<string> oldDepthVisitedTilings = new HashSet<string>();
            HashSet<string> currentDepthVisitedTilings = new HashSet<string>();
            HashSet<string> nextDepthVisitedTilings = new HashSet<string>();

            //HashSet<string> visitedTilings = new HashSet<string>();

            HashSet<string> posVisitedParents = new HashSet<string>();
            HashSet<string> negVisitedParents = new HashSet<string>();

            var positive = new List<CCInfo>();
            var negative = new List<CCInfo>();

            //Stack<string> tilingsCC = new Stack<string>();
            //tilingsCC.Push(tiling);

            currentDepthVisitedTilings.Add(tiling);

            //visitedTilings.Add(tiling);

            //Used for logging purposes
            long numTilings = 0;
            long milestone, step = milestone = 500000;
            long depth = -1;
            //Stopwatch stopWatch = Stopwatch.StartNew();

            using (StreamWriter writer = new StreamWriter(filename))
            {

                while (currentDepthVisitedTilings.Count > 0)
                {
                    depth++;
                    if (numTilings > milestone)
                    {
                        Console.WriteLine(DateTime.Now);
                        Console.WriteLine("Reached {0} tilings", numTilings);
                        Console.WriteLine("Current depth: {0}\n", depth);
                        milestone += step;
                    }

                    writer.WriteLine("{0} - Depth = {1}, Number of tilings = {2}", DateTime.Now, depth, currentDepthVisitedTilings.Count);

                    foreach (string til in currentDepthVisitedTilings)
                    {
                        numTilings++;
                        //string til = tilingsCC.Pop();
                        Region tReg = Region.FromString(til);

                        var trits = tReg.FindTrits();

                        foreach (Trit trit in trits)
                        {
                            string tritted = tReg.GetTrittedString(trit);
                            var info = componentsInfo.GetConnectedComponent(tritted);

                            if (trit.IsPositive())
                            {
                                if (posVisitedParents.Add(info.Item1))
                                {
                                    positive.Add(info);
                                    writer.WriteLine("\r\n{0} - Found new positive trit: {1}\r\n", DateTime.Now, info.Item2);
                                }
                            }
                            else
                            {
                                if (negVisitedParents.Add(info.Item1))
                                {
                                    negative.Add(info);
                                    writer.WriteLine("\r\n{0} - Found new negative trit: {1}\r\n", DateTime.Now, info.Item2);
                                }
                            }

                        }



                        List<string> flips = tReg.FindFlips();

                        foreach (string t in flips)
                        {
                            if (!oldDepthVisitedTilings.Contains(t) && !currentDepthVisitedTilings.Contains(t))
                            {
                                nextDepthVisitedTilings.Add(t);
                                //tilingsCC.Push(t);
                            }
                        }
                    }

                    oldDepthVisitedTilings = currentDepthVisitedTilings;
                    currentDepthVisitedTilings = nextDepthVisitedTilings;
                    nextDepthVisitedTilings = new HashSet<string>();

                }

                Console.WriteLine("{0} - Finished full search", DateTime.Now);
                writer.WriteLine("{0} - Finished full search", DateTime.Now);
                //stopWatch.Stop();

                writer.WriteLine("Tiling: {0}\n", tiling);
                writer.WriteLine("------- Positive Trits ------");
                foreach (var info in positive)
                {
                    writer.WriteLine("Parent: {0}\r\nIndex: {1}\r\nNumber of Elems: {2}\r\n", info.Item1, info.Item2, info.Item3);
                }

                writer.WriteLine("------ Negative Trits -----");

                foreach (var info in negative)
                {
                    writer.WriteLine("Parent: {0}\r\nIndex: {1}\r\nNumber of Elems: {2}\r\n", info.Item1, info.Item2, info.Item3);
                }
            }


        }

        #region Tests of specific functionalities

        static void ToAndFromTilingStringTest()
        {
            StringPair sp = new StringPair("IBIB,RLDI,IIUI,IBIB;DFDF,UDUD,DUDU,UFUF", "IBIB,RLDI,IIUI,IBIB;DFDF,UDUD,DUDU,UFUF");
            var tiling = sp.ToTilingString();
            Console.WriteLine(tiling);
            sp = StringPair.FromTilingString(tiling);
            Console.WriteLine(sp);

            sp = new StringPair("IBID,IIDU,IDUI,IUIB;DFRL,URLD,DRLU,URLF", "IBID,IIBU,IRLI,IBIB;DFRL,UDFD,DUDU,UFUF");
            tiling = sp.ToTilingString();
            Console.WriteLine(tiling);
            sp = StringPair.FromTilingString(tiling);
            Console.WriteLine(sp);
        }

        static bool OppositeTest()
        {
            DimerDirection[] possibleValues = {DimerDirection.Right,DimerDirection.Left,
                                                  DimerDirection.Down,DimerDirection.Up,
                                                  DimerDirection.Back,DimerDirection.Front};

            DimerDirection[] oppositeValues = {DimerDirection.Left,DimerDirection.Right,
                                                  DimerDirection.Up,DimerDirection.Down,
                                                  DimerDirection.Front,DimerDirection.Back};

            bool success = true;

            for (int i = 0; i < 6; i++)
            {
                if (Region.Opposite(possibleValues[i]) != oppositeValues[i])
                {
                    Console.WriteLine("Opposite Fail:\nDirection = {0}, Given opposite = {1}, Actual opposite = {2}",
                        possibleValues[i], Region.Opposite(possibleValues[i]), oppositeValues[i]);
                    success = false;
                }
            }

            Console.WriteLine("Empty and Inaccessible Behavior:");
            Console.WriteLine("Empty opposite: {0}", Region.Opposite(DimerDirection.Empty));
            Console.WriteLine("Inaccessible opposite: {0}", Region.Opposite(DimerDirection.Inaccessible));

            return success;


        }

        static bool IncrementTest()
        {
            Region r = new Region(new BoxDimensions(4, 6, 3));

            DimerDirection[] possibleValues = {DimerDirection.Right,DimerDirection.Left,
                                                  DimerDirection.Down,DimerDirection.Up,
                                                  DimerDirection.Back,DimerDirection.Front};

            int[] increments = { 1, -1, 4, -4, 24, -24 };

            bool success = true;

            for (int i = 0; i < 6; i++)
            {
                if (r.IncrementToNeighborInDirection(possibleValues[i]) != increments[i])
                {
                    Console.WriteLine("Increment Fail:\nDirection = {0}, Given increment = {1}, Actual increment = {2}",
                        possibleValues[i], r.IncrementToNeighborInDirection(possibleValues[i]), increments[i]);
                    success = false;
                }
            }

            try
            {
                r.IncrementToNeighborInDirection(DimerDirection.Empty);
                Console.WriteLine("Method did not throw exception with Empty parameter");
                success = false;
            }
            catch (Exception ex)
            {
                if (!(ex is ArgumentOutOfRangeException))
                {
                    Console.WriteLine("Exception thrown with Empty parameter is not ArgumentOutOfRangeException");
                    success = false;
                }

            }

            try
            {
                r.IncrementToNeighborInDirection(DimerDirection.Inaccessible);
                Console.WriteLine("Method did not throw exception with Inaccessible parameter");
                success = false;
            }
            catch (Exception ex)
            {
                if (!(ex is ArgumentOutOfRangeException))
                {
                    Console.WriteLine("Exception thrown with Inaccessible parameter is not ArgumentOutOfRangeException");
                    success = false;
                }

            }

            return success;
        }

        static void TritTest()
        {
            Region r = Region.FromString("RLD,DBU,URL;DRL,UFD,RLU");

            Trit trit1 = new Trit(0, DimerDirection.Right, DimerDirection.Back);
            Trit trit2 = new Trit(4, DimerDirection.Back, DimerDirection.Down);

            Console.WriteLine(r.GetTrittedString(trit1));

            Console.WriteLine(r.GetTrittedString(trit2));

            r = Region.FromString("DRL,UBD,RLU;RLD,DFU,URL");

            Trit trit3 = new Trit(4, DimerDirection.Back, DimerDirection.Down);

            Console.WriteLine(r.GetTrittedString(trit3));





        }

        static void FindTrittedTIlingsTest()
        {
            //Region r = Region.FromString("RLD,DBU,URL;DRL,UFD,RLU");

            Region r = Region.FromString("DRL,UBD,RLU;RLD,DFU,URL");

            List<string> tritsFromRegion = r.FindTrittedTilings();

            StringPair sp = new StringPair("IDID,BUBU,RLRL,IBIB;RLRL,FDFD,DUDU,UFUF", "IBIB,RLRL,BDBD,IUIU;DFDF,UDUD,FUFU,RLRL");

            Console.WriteLine(Region.FindFCC(sp.ToTilingString()).Count);

            Region r2 = Region.FromString(sp.ToTilingString());

            List<string> trittedChoices = r2.FindTrittedTilings();

            foreach (string t in trittedChoices)
            {
                Console.WriteLine("{0}", Region.FindFCC(t).Count);
            }


        }

        static void PositiveTritTest()
        {
            DimerDirection[] directions = new DimerDirection[] { DimerDirection.Right, DimerDirection.Back, DimerDirection.Down };

            foreach (DimerDirection d in directions)
                foreach (DimerDirection d2 in directions)
                {
                    if (d == d2)
                        continue;

                    Trit t = new Trit(0, d, d2, false);
                    Console.WriteLine("Initial Direction: {0}\nFinal Direction: {1}\n Positive: {2}", t.InitialDirection, t.FinalDirection, t.IsPositive());
                }

            Trit t2 = new Trit(0, DimerDirection.Down, DimerDirection.Back, true);
            Trit t3 = new Trit(0, DimerDirection.Back, DimerDirection.Down, true);

            Console.WriteLine("Initial Direction: {0}\nFinal Direction: {1}\n Positive: {2}", t2.InitialDirection, t2.FinalDirection, t2.IsPositive());
            Console.WriteLine("Initial Direction: {0}\nFinal Direction: {1}\n Positive: {2}", t3.InitialDirection, t3.FinalDirection, t3.IsPositive());
        }

        #endregion


    }


}
