﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using System.IO;
//using StringPair = System.Tuple<string, string>;

namespace TilingsConsole
{

    public enum DimerDirection
    {
        Inaccessible = -1,
        Empty = 0,
        Right = 2,
        Down = 3,
        Back = 4,
        Left = -2,
        Up = -3,
        Front = -4,
    }

    public class BoxDimensions
    {
        public int Width { set; get; }
        public int Height { set; get; }
        public int Depth { set; get; }

        public BoxDimensions(int width, int height, int depth)
        {
            Width = width;
            Height = height;
            Depth = depth;
        }
    }

    public class Region
    {

        #region Parameters and properties

        public int Width { private set; get; }
        public int Height { private set; get; }
        public int Depth { private set; get; }

        public DimerDirection[] RegionArray;
        public int Length { get { return RegionArray.Length; } }





        public List<string> TilingsList;

        long numTilings;
        public long NumTilings { get { return numTilings; } }

        #endregion

        #region Constructors

        public Region(DimerDirection[] regionArray, int widthParam, int heightParam, int depthParam)
        {
            RegionArray = regionArray;
            Width = widthParam;
            Height = heightParam;
            Depth = depthParam;

            if (RegionArray.Length != Width * Height * Depth)
            {
                Console.WriteLine("Warning: Region dimensions do not match");
            }


        }

        public Region(BoxDimensions boxDim)
        {
            RegionArray = new DimerDirection[boxDim.Width * boxDim.Height * boxDim.Depth];
            Width = boxDim.Width;
            Height = boxDim.Height;
            Depth = boxDim.Depth;
        }

        #endregion


        #region Indexers and Related Helper Methods

        public DimerDirection this[int pos]
        {
            get
            {
                return RegionArray[pos];
            }
            set
            {
                RegionArray[pos] = value;
            }
        }

        public DimerDirection this[int i, int j, int k]
        {
            get
            {
                int pos = Pos(i, j, k);
                return this[pos];
            }
            set
            {
                int pos = Pos(i, j, k);
                this[pos] = value;
            }
        }

        public BoxDimensions GetBoxDimensions() { return new BoxDimensions(Width, Height, Depth); }



        public void Empty()
        {
            for (int pos = 0; pos < Length; pos++)
            {
                if (RegionArray[pos] != DimerDirection.Inaccessible)
                    RegionArray[pos] = DimerDirection.Empty;
            }
        }

        public bool IsWhite(int pos)
        {
            int[] indexes = Indexes(pos);

            int i = indexes[0];
            int j = indexes[1];
            int k = indexes[2];

            return ((i + j + k) % 2 == 0);
        }

        public int[] Indexes(int pos)
        {
            int WidthTimesHeight = Width * Height;

            int i = pos % Width;
            int j = (pos / Width) % Height;
            int k = pos / WidthTimesHeight;

            return new int[3] { i, j, k };

        }

        public void IndexVariables(int pos, out int i, out int j, out int k)
        {
            int[] indexes = Indexes(pos);
            i = indexes[0];
            j = indexes[1];
            k = indexes[2];

        }

        public int Pos(int i, int j, int k)
        {
            return i + Width * j + Width * Height * k;
        }

        public bool IsPosInRegion(int pos)
        {
            return ((pos >= 0) && (pos < Length) && this[pos] != DimerDirection.Inaccessible);
        }
        #endregion

        #region DimerDirection Auxiliary Methods

        public static DimerDirection Opposite(DimerDirection direction)
        {
            int directionAsInt = (int)direction;
            directionAsInt = -directionAsInt;

            return (DimerDirection)directionAsInt;
        }

        public int IncrementToNeighborInDirection(DimerDirection direction)
        {
            if ((direction == DimerDirection.Empty) || (direction == DimerDirection.Inaccessible))
            {
                throw new ArgumentOutOfRangeException("direction", direction, "direction parameter cannot be Empty or Inaccessible");
            }

            int increment;

            switch (direction)
            {
                case DimerDirection.Right:
                    increment = 1;
                    break;
                case DimerDirection.Down:
                    increment = Width;
                    break;
                case DimerDirection.Back:
                    increment = Width * Height;
                    break;
                default:
                    increment = -IncrementToNeighborInDirection(Opposite(direction));
                    break;
            }

            return increment;
        }

        public void AddDimer(int initPos, DimerDirection direction)
        {
            this[initPos] = direction;
            this[initPos + IncrementToNeighborInDirection(direction)] = Opposite(direction);
        }

        #endregion

        #region Converting String <-> Region

        public override string ToString()
        {
            StringBuilder tiledRegion = new StringBuilder(Length + Height * Depth);

            for (int i = 0; i < Length; i++)
            {
                if (i > 0 && i < Length)
                {
                    if (i % (Width * Height) == 0)
                        tiledRegion.Append(';');
                    else if (i % Width == 0)
                        tiledRegion.Append(',');
                }

                char c = 'E';
                switch (this[i])
                {
                    case DimerDirection.Back:
                        c = 'B';
                        break;
                    case DimerDirection.Down:
                        c = 'D';
                        break;
                    case DimerDirection.Empty:
                        c = 'E';
                        break;
                    case DimerDirection.Front:
                        c = 'F';
                        break;
                    case DimerDirection.Inaccessible:
                        c = 'I';
                        break;
                    case DimerDirection.Left:
                        c = 'L';
                        break;
                    case DimerDirection.Right:
                        c = 'R';
                        break;
                    case DimerDirection.Up:
                        c = 'U';
                        break;


                }

                tiledRegion.Append(c);

            }

            return tiledRegion.ToString();
        }

        public static Region FromString(string tiling)
        {
            //Uses the new implementation of tiling strings

            int depth = tiling.Count(c => c == ';') + 1;

            int height = 1, width = 0;
            int index = 0;
            for (index = 0; index < tiling.Length; index++)
            {
                char c = tiling[index];

                if (c == ';')
                {
                    break;
                }
                if (c == ',')
                {
                    height++;
                    index++;
                    break;
                }

                width++;
            }

            for (; index < tiling.Length; index++)
            {
                char c = tiling[index];
                if (c == ',')
                    height++;
                if (c == ';')
                    break;
            }
            // First, width, height and depth were inferred from the string.
            // Now we proceed to building the region

            return FromString(tiling, new BoxDimensions(width, height, depth));


        }

        public static Region FromString(string tiling, BoxDimensions boxDim)
        {
            int width = boxDim.Width, height = boxDim.Height, depth = boxDim.Depth;

            DimerDirection[] regArray = new DimerDirection[width * height * depth];

            int j = 0;
            for (int i = 0; i < tiling.Length; i++)
            {
                DimerDirection direction = DimerDirection.Empty;

                switch (tiling[i])
                {
                    case 'B':
                        direction = DimerDirection.Back;
                        break;
                    case 'D':
                        direction = DimerDirection.Down;
                        break;
                    case 'E':
                        direction = DimerDirection.Empty;
                        break;
                    case 'F':
                        direction = DimerDirection.Front;
                        break;
                    case 'I':
                        direction = DimerDirection.Inaccessible;
                        break;
                    case 'L':
                        direction = DimerDirection.Left;
                        break;
                    case 'R':
                        direction = DimerDirection.Right;
                        break;
                    case 'U':
                        direction = DimerDirection.Up;
                        break;
                    default:
                        continue;


                }

                regArray[j] = direction;
                j++;


            }

            return new Region(regArray, width, height, depth);

        }

        /// <summary>
        /// Returns a string representing the region where the current tiling lies,
        /// only empty
        /// </summary>
        /// <param name="tiling">A string representing a tiling</param>
        /// <returns>A string representing the (empty) region where the tiling lies</returns>
        public static string EmptyString(string tiling)
        {
            return string.Join("", tiling.Select(x => ("RLBFUD".Contains(x) ? 'E' : x)));
        }
        #endregion

        #region Converting to Box

        public bool IsEvenDepthBox()
        {
            if (Depth % 2 != 0)
                return false;

            bool inaccessibleSquareExists = Array.Exists<DimerDirection>(RegionArray, st => (st == DimerDirection.Inaccessible));
            return !inaccessibleSquareExists;
        }


        #endregion

        #region Auxiliary tiling methods

        public bool TryTileRegion(int startingPosition)
        {
            for (int pos = startingPosition; pos < Length; pos++)
            {
                if (this[pos] != DimerDirection.Empty)
                    continue;

                int[] indexes = Indexes(pos);
                int i = indexes[0];
                int j = indexes[1];
                int k = indexes[2];

                if ((i + 1 < Width) && (this[i + 1, j, k] == DimerDirection.Empty))
                {
                    this[i, j, k] = DimerDirection.Right;
                    this[i + 1, j, k] = DimerDirection.Left;
                }
                else if ((j + 1 < Height) && (this[i, j + 1, k] == DimerDirection.Empty))
                {
                    this[i, j, k] = DimerDirection.Down;
                    this[i, j + 1, k] = DimerDirection.Up;
                }
                else if ((k + 1 < Depth) && (this[i, j, k + 1] == DimerDirection.Empty))
                {
                    this[i, j, k] = DimerDirection.Back;
                    this[i, j, k + 1] = DimerDirection.Front;
                }
                else
                {
                    return IsTiled(); //We got cornered, if it isn't already tiled then we can't tile the region
                }
            }

            return true;
        }

        DimerDirection RemoveLastTileBefore(ref int pos)
        {
            while (pos >= 0)
            {
                int i, j, k;
                IndexVariables(pos, out i, out j, out k);

                switch (this[pos])
                {
                    case DimerDirection.Right:
                        this[i, j, k] = this[i + 1, j, k] = DimerDirection.Empty;
                        return DimerDirection.Right;

                    case DimerDirection.Down:
                        this[i, j, k] = this[i, j + 1, k] = DimerDirection.Empty;
                        return DimerDirection.Down;

                    case DimerDirection.Back:
                        this[i, j, k] = this[i, j, k + 1] = DimerDirection.Empty;
                        return DimerDirection.Back;


                }

                pos--;
            }

            return DimerDirection.Empty;

        }

        int PrepareNextTiling() //returns the starting position
        {

            int pos = Length - 1;
            while (pos >= 0)
            {
                DimerDirection direction = RemoveLastTileBefore(ref pos);

                if (direction == DimerDirection.Empty)
                    return -1;

                int i, j, k;
                IndexVariables(pos, out i, out j, out k);

                if (direction == DimerDirection.Back)
                    continue;

                if (k == Depth - 1)
                {
                    if ((j == Height - 1) || (direction == DimerDirection.Down) || (this[i, j + 1, k] != DimerDirection.Empty))
                        continue;

                    this[i, j, k] = DimerDirection.Down;
                    this[i, j + 1, k] = DimerDirection.Up;

                }
                else if ((direction == DimerDirection.Down) || (j == Height - 1) || this[i, j + 1, k] != DimerDirection.Empty)
                {
                    if (this[i, j, k + 1] == DimerDirection.Empty)
                    {
                        this[i, j, k] = DimerDirection.Back;
                        this[i, j, k + 1] = DimerDirection.Front;
                    }
                    else
                    {
                        continue;
                    }
                }
                else //direction == Right, j <= Height - 1 and this[i,j+1,k] == Empty
                {
                    this[i, j, k] = DimerDirection.Down;
                    this[i, j + 1, k] = DimerDirection.Up;
                }


                return pos + 1;
            }

            return -1;

        }


        public bool IsTiled()
        {
            bool emptySquareExists = Array.Exists<DimerDirection>(RegionArray, st => (st == DimerDirection.Empty));
            return !emptySquareExists;
        }

        #endregion

        #region Find One Tiling and Embed Into Box

        public bool TileRegionIfPossible()
        {
            int pos = 0;

            while (pos >= 0)
            {
                if (TryTileRegion(pos))
                    return true;

                pos = PrepareNextTiling();
            }

            return false;

        }

        public string EmbedIntoBox(BoxDimensions boxDim)
        {
            //These represent that lowest coordinate of the current region as it's embeded into boxDim
            int x = (boxDim.Width - Width) >> 1;
            int y = (boxDim.Height - Height) >> 1;
            int z = (boxDim.Depth - Depth) >> 1;

            Region box = new Region(boxDim);

            for (int i = 0; i < Width; i++)
                for (int j = 0; j < Height; j++)
                    for (int k = 0; k < Depth; k++)
                    {
                        box[x + i, y + j, z + k] = DimerDirection.Inaccessible;
                    }

            if (box.TileRegionIfPossible())
            {
                for (int i = 0; i < Width; i++)
                    for (int j = 0; j < Height; j++)
                        for (int k = 0; k < Depth; k++)
                        {
                            box[x + i, y + j, z + k] = this[i, j, k];
                        }

                return box.ToString();

            }
            else
                return null;

        }

        #endregion

        #region Counting methods

        public long CountTilings()
        {

            numTilings = 0;
            int pos = 0;

            while (pos >= 0)
            {
                if (TryTileRegion(pos))
                {
                    numTilings++;
                }


                pos = PrepareNextTiling();

            }

            return numTilings;
        }

        #endregion

        #region Flip Connected Component Methods

        public Dictionary<string, UnionFindNode<string>> FindFlipConnectedComponents()
        {
            Dictionary<string, UnionFindNode<string>> connectedComponentsDic = new Dictionary<string, UnionFindNode<string>>(new StringEqualityComparer());

            //numTilings = 0;
            int pos = 0;

            //stopwatch.Start();
            while (pos >= 0)
            {
                if (TryTileRegion(pos))
                {
                    string currentRegion = this.ToString();
                    if (!connectedComponentsDic.ContainsKey(currentRegion))
                    {
                        UnionFindNode<string> currentNode = new UnionFindNode<string>(currentRegion);
                        connectedComponentsDic.Add(currentRegion, currentNode);
                        List<string> flipsList = FindFlips();

                        foreach (string tiling in flipsList)
                        {
                            UnionFindNode<string> tilNode;
                            if (!connectedComponentsDic.TryGetValue(tiling, out tilNode))
                            {
                                tilNode = new UnionFindNode<string>(tiling);
                                connectedComponentsDic.Add(tiling, tilNode);

                            }

                            UnionFindNode.Union(currentNode, tilNode);
                        }
                    }
                    else
                    {
                        List<string> flipsList = FindFlips();

                        foreach (string tiling in flipsList)
                        {
                            UnionFindNode<string> tilNode;
                            if (connectedComponentsDic.TryGetValue(tiling, out tilNode))
                            {
                                UnionFindNode.Union(connectedComponentsDic[currentRegion], tilNode);
                            }
                        }
                    }

                }


                pos = PrepareNextTiling();

            }

            //stopwatch.Stop();
           // Console.WriteLine("{0} - Found all {1} tilings", DateTime.Now, numTilings);

            return connectedComponentsDic;
        }

        public static HashSet<string> FindFCC(string tiling)
        {
            HashSet<string> visitedTilings = new HashSet<string>();

            Stack<string> tilingsCC = new Stack<string>();
            tilingsCC.Push(tiling);
            visitedTilings.Add(tiling);

            while (tilingsCC.Count > 0)
            {
                string til = tilingsCC.Pop();
                Region tReg = Region.FromString(til);
                List<string> flips = tReg.FindFlips();

                foreach (string t in flips)
                {
                    if (visitedTilings.Add(t))
                    {
                        tilingsCC.Push(t);
                    }
                }

            }

            return visitedTilings;
        }

        #endregion

        #region Flip methods

        /// <summary>
        /// Finds all flips from the current region, and returns a list of all the tilings that can be reached after one flip.
        /// </summary>
        /// <returns>A list of tiling strings.</returns>
        public List<string> FindFlips()
        {
            List<string> flipsList = new List<string>();

            for (int pos = 0; pos < Length; pos++)
            {
                DimerDirection direction = this[pos];
                if (direction == DimerDirection.Right || direction == DimerDirection.Down || direction == DimerDirection.Back)
                {
                    int i, j, k;
                    IndexVariables(pos, out i, out j, out k);

                    int initPos1 = pos, initPos2;


                    if ((i + 1 < Width) && this[i + 1, j, k] == direction)
                    {
                        initPos2 = pos + 1;

                        flipsList.Add(GetFlip(initPos1, initPos2, DimerDirection.Right, DimerDirection.Left));


                    }

                    if ((j + 1 < Height) && this[i, j + 1, k] == direction)
                    {
                        initPos2 = pos + Width;

                        flipsList.Add(GetFlip(initPos1, initPos2, DimerDirection.Down, DimerDirection.Up));
                    }

                    if ((k + 1 < Depth) && this[i, j, k + 1] == direction)
                    {
                        initPos2 = pos + Width * Height;

                        flipsList.Add(GetFlip(initPos1, initPos2, DimerDirection.Back, DimerDirection.Front));
                    }


                }

            }


            return flipsList;
        }

        private string GetFlip(int initPos1, int initPos2, DimerDirection newDirectionInit, DimerDirection newDirectionFinal)
        {
            DimerDirection oldDirectionInit = this[initPos1];


            int increment = -1;

            switch (oldDirectionInit)
            {
                case DimerDirection.Right:
                    increment = 1;
                    break;
                case DimerDirection.Down:
                    increment = Width;
                    break;
                case DimerDirection.Back:
                    increment = Width * Height;
                    break;
            }

            if (increment < 0)
            {
                throw new Exception("Direction was not Right, Down or Back");
            }

            DimerDirection oldDirectionFinal = this[initPos1 + increment];


            this[initPos1] = this[initPos1 + increment] = newDirectionInit;
            this[initPos2] = this[initPos2 + increment] = newDirectionFinal;

            string flippedRegion = this.ToString();


            this[initPos1] = this[initPos2] = oldDirectionInit;
            this[initPos1 + increment] = this[initPos2 + increment] = oldDirectionFinal;

            return flippedRegion;

        }


        #endregion

        #region Trit Methods

        public static void FindTritsFromCC(string tiling, out HashSet<string> posTrits, out HashSet<string> negTrits)
        {
            HashSet<string> visitedTilings = new HashSet<string>();

            posTrits = new HashSet<string>();
            negTrits = new HashSet<string>();

            Stack<string> tilingsCC = new Stack<string>();
            tilingsCC.Push(tiling);
            visitedTilings.Add(tiling);

            while (tilingsCC.Count > 0)
            {
                string til = tilingsCC.Pop();
                Region tReg = Region.FromString(til);

                var trits = tReg.FindTrits();

                foreach (Trit trit in trits)
                {
                    string tritted = tReg.GetTrittedString(trit);
                    if (trit.IsPositive())
                    {
                        posTrits.Add(tritted);
                    }
                    else
                    {
                        negTrits.Add(tritted);
                    }
                }



                List<string> flips = tReg.FindFlips();

                foreach (string t in flips)
                {
                    if (visitedTilings.Add(t))
                    {
                        tilingsCC.Push(t);
                    }
                }

            }

        }

        public List<Trit> FindTrits()
        {
            List<Trit> trits = new List<Trit>();

            for (int pos = 0; pos < Length; pos++)
            {
                DimerDirection direction = this[pos];

                List<DimerDirection> validDirections = new List<DimerDirection>() { DimerDirection.Right, DimerDirection.Down, DimerDirection.Back };

                if (!validDirections.Contains(direction))
                    continue;

                validDirections.Remove(direction);

                for (int i = 0; i < 2; i++)
                {
                    var finalDirection = validDirections[i];
                    var otherDirection = validDirections[1 - i];

                    int secondPos = pos + IncrementToNeighborInDirection(finalDirection);

                    if (!IsPosInRegion(secondPos))
                        continue;

                    int thirdPos = pos + IncrementToNeighborInDirection(direction) + IncrementToNeighborInDirection(this[secondPos]);


                    if (!IsPosInRegion(thirdPos))
                        continue;

                    bool tritExists = (this[secondPos] == otherDirection) && (this[thirdPos] == finalDirection);

                    if (tritExists)
                    {
                        trits.Add(new Trit(pos, direction, finalDirection, false));
                        continue;
                    }

                    if (otherDirection == DimerDirection.Right)
                    {
                        tritExists = (this[secondPos] == DimerDirection.Left) && (this[thirdPos] == finalDirection);
                        if (tritExists)
                        {
                            trits.Add(new Trit(pos, direction, finalDirection, true));
                        }
                    }



                }


            }

            return trits;
        }


        public List<string> FindTrittedTilings()
        {
            return FindTrits().Select(x => GetTrittedString(x)).ToList();
        }

        void MakeTrit(Trit trit)
        {
            DimerDirection initDirection = trit.InitialDirection;
            DimerDirection finalDirection = trit.FinalDirection;
            DimerDirection otherDirection = this[trit.InitPos + IncrementToNeighborInDirection(finalDirection)];

            int incrementForInitDirection = IncrementToNeighborInDirection(initDirection);
            int incrementForFinalDirection = IncrementToNeighborInDirection(finalDirection);
            int incrementForTheOtherDirection = IncrementToNeighborInDirection(otherDirection);

            int initPos = trit.InitPos;
            int secondPos = initPos + incrementForInitDirection;
            int thirdPos = initPos + incrementForFinalDirection + incrementForTheOtherDirection;

            AddDimer(initPos, finalDirection);
            AddDimer(secondPos, otherDirection);
            AddDimer(thirdPos, initDirection);

        }

        void UndoTrit(Trit trit)
        {
            Trit reverseTrit = new Trit(trit.InitPos, trit.FinalDirection, trit.InitialDirection);
            MakeTrit(reverseTrit);
        }

        public string GetTrittedString(Trit trit)
        {
            MakeTrit(trit);
            string trittedRegion = ToString();
            UndoTrit(trit);
            return trittedRegion;
        }

        #endregion

        #region Depth Two Invariant

        /// <summary>
        /// Calculates the depth two invariant (a polynomial), if the region is of the form D x [0,2]:
        /// see http://arxiv.org/abs/1404.6509 .
        /// </summary>
        /// <returns>A string representing the polynomial.</returns>
        public string DepthTwoEqualInvariant()
        {//For the time being, the invariant only works for regions of the form D x [0,2], where D is a 2-dimensional figure 

            if (Depth != 2)
                return "";

            int WidthTimesHeight = Width * Height;
            int mindim = (Width < Height ? Width : Height);
            int maxCycles = (mindim - 1) >> 1;

            int[] invariant = new int[mindim]; //There are at most (mindim - 1) >> 1 cycles surrounding the center 

            Dictionary<DimerDirection, int> directions = new Dictionary<DimerDirection, int>();
            directions[DimerDirection.Up] = 1;
            directions[DimerDirection.Down] = -1;
            directions[DimerDirection.Back] = directions[DimerDirection.Front] = directions[DimerDirection.Right] =
                directions[DimerDirection.Left] = directions[DimerDirection.Inaccessible] = 0;

            for (int i = 0; i < WidthTimesHeight; i++)
            {
                int count = 0;
                if (this[i] == DimerDirection.Back)
                {
                    for (int j = i + 1; j % Width != 0; j++)
                    {
                        bool IsWhiteTile = IsWhite(j);
                        int diff = directions[this[j]] - directions[this[j + WidthTimesHeight]];
                        count += (IsWhiteTile ? diff : -diff);
                    }

                    count >>= 1;

                    invariant[maxCycles + count] += (!IsWhite(i) ? 1 : -1);
                }
            }


            return string.Join(" + ", invariant.Select((x,i) => string.Format((i != maxCycles ? "{1}q^{{{0}}}" : "{1}"), i - maxCycles, x)).Where(x => x[0] != '0').Reverse());

        }

        #endregion

        public double Twist()
        {

            double totalEffect = 0;
            Dictionary<DimerDirection, double> rightLeft = new Dictionary<DimerDirection, double>() { { DimerDirection.Right, 1 }, { DimerDirection.Left, -1 } };
            Dictionary<DimerDirection, double> upDown = new Dictionary<DimerDirection, double>() { { DimerDirection.Up, 1 }, { DimerDirection.Down, -1 } };

            for (int pos = 0; pos < Length; pos++)
            {
                double effect = 0.25;
                var c = this[pos];
                var dic = upDown;
                var otherDic = rightLeft;
                if (upDown.ContainsKey(c))
                {
                    effect = -effect;
                    dic = rightLeft;
                    otherDic = upDown;
                    
                }
                else if (!rightLeft.ContainsKey(c))
                {
                    continue;
                }

                if (otherDic[c] < 0)
                    effect = -effect;

                if (!IsWhite(pos))
                    effect = -effect;



                int i, j, k;
                IndexVariables(pos, out i, out j, out k);

                for (int w = k + 1; w < Depth; w++)
                {
                    double thisEffect = effect;
                    var c2 = this[i, j, w];
                    if (!dic.ContainsKey(c2))
                    {
                        continue;
                    }

                    if (dic[c2] < 0)
                        thisEffect = -thisEffect;

                    if ((i + j + w) % 2 != 0)
                        thisEffect = -thisEffect;

                    totalEffect += thisEffect;
                }


            }

            return totalEffect;
        }


        #region Counting methods for boxes

        /// <summary>
        /// Counts the number of tilings of a box. Does not use parallel computing.
        /// </summary>
        /// <param name="boxDim">The dimension of the box.</param>
        /// <returns>The number of tilings.</returns>
        public static long SmartCount(BoxDimensions boxDim)
        {

            int width = boxDim.Width;
            int height = boxDim.Height;
            int depth = boxDim.Depth;

            if (depth % 2 != 0)
            {
                if (width % 2 == 0)
                    return SmartCount(new BoxDimensions(boxDim.Depth, boxDim.Height, boxDim.Width));
                else if (height % 2 == 0)
                    return SmartCount(new BoxDimensions(boxDim.Width, boxDim.Depth, boxDim.Height));
                else
                    return 0;
            }
            Console.WriteLine("{3}: Started counting tilings of {0}x{1}x{2} box", width, height, depth, DateTime.Now);
            int halfDepth = depth >> 1;
            int WidthHeigth = width * height;
            int halfWidthHeight = (halfDepth % 2 == 0 ? (WidthHeigth >> 1) : (WidthHeigth + 1) >> 1);

            var baseRegion = new Region(new BoxDimensions(width, height, halfDepth));

            long sum = 0;

            int[] chosenPositions = new int[halfWidthHeight + 1]; //The last one is an auxiliary construction
            for (int i = 0; i < halfWidthHeight; i++)
            {
                chosenPositions[i] = i;
            }

            chosenPositions[halfWidthHeight] = WidthHeigth + 10;

            int regionsCalc = 0;
            Console.WriteLine("Number of regions to calc: {0}", MathConvenience.nChoosek(WidthHeigth, halfWidthHeight).ToString("N0"));



            while (chosenPositions[halfWidthHeight - 1] < WidthHeigth)
            {

                int i = 0;
                for (int chPos = 0; chPos < halfWidthHeight; chPos++)
                {

                    while (i < chosenPositions[chPos])
                    {
                        baseRegion[i] = (baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);
                        i++;
                    }


                    baseRegion[i] = (!baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);
                    i++;

                }

                for (; i < WidthHeigth; i++)
                    baseRegion[i] = (baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);

                long n = baseRegion.CountTilings();


                sum += n * n;

                for (int chPos = 0; chPos < halfWidthHeight; chPos++)
                {
                    if (chosenPositions[chPos + 1] > (chosenPositions[chPos] + 1))
                    {
                        chosenPositions[chPos]++;
                        for (int ind = 0; ind < chPos; ind++)
                            chosenPositions[ind] = ind;

                        break;
                    }

                }

                regionsCalc++;





            }

            Console.WriteLine("Regions: {0}", regionsCalc);
            Console.WriteLine("Tilings: {0}", sum.ToString("N0"));
            Console.WriteLine("Finished: {0}", DateTime.Now);
            return sum;

        }

        /// <summary>
        /// Counts the number of tilings of a box. Uses parallel computing.
        /// </summary>
        /// <param name="boxDim">The dimension of the box.</param>
        /// <returns>The number of tilings.</returns>
        public static long ParallelSmartCount(BoxDimensions boxDim)
        {

            int width = boxDim.Width;
            int height = boxDim.Height;
            int depth = boxDim.Depth;

            if (depth % 2 != 0)
            {
                if (width % 2 == 0)
                    return ParallelSmartCount(new BoxDimensions(depth, height, width));
                else if (height % 2 == 0)
                    return ParallelSmartCount(new BoxDimensions(width, depth, height));
                else
                    return 0;
            }
            Console.WriteLine("{3}: Started counting tilings of {0}x{1}x{2} box", width, height, depth, DateTime.Now);
            int halfDepth = depth >> 1;
            int WidthHeigth = width * height;
            int halfWidthHeight = (halfDepth % 2 == 0 ? (WidthHeigth >> 1) : (WidthHeigth + 1) >> 1);



            long sum = 0;

            int[] chosenPositions = new int[halfWidthHeight + 1]; //The last one is an auxiliary construction
            for (int i = 0; i < halfWidthHeight; i++)
            {
                chosenPositions[i] = i;
            }

            chosenPositions[halfWidthHeight] = WidthHeigth + 10;

            int regionsCalc = 0;
            long totalRegions = MathConvenience.nChoosek(WidthHeigth, halfWidthHeight);
            Console.WriteLine("Number of regions to calc: {0}", totalRegions.ToString("N0"));

            Task<long>[] tasks = new Task<long>[totalRegions];

            while (chosenPositions[halfWidthHeight - 1] < WidthHeigth)
            {
                var baseRegion = new Region(new BoxDimensions(width, height, halfDepth));
                int i = 0;
                for (int chPos = 0; chPos < halfWidthHeight; chPos++)
                {

                    while (i < chosenPositions[chPos])
                    {
                        baseRegion[i] = (baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);
                        i++;
                    }


                    baseRegion[i] = (!baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);
                    i++;

                }

                for (; i < WidthHeigth; i++)
                    baseRegion[i] = (baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);

                tasks[regionsCalc] = Task<long>.Factory.StartNew((region) =>
                {
                    Region r = region as Region;
                    return r.CountTilings();
                }, baseRegion);


                //sum += n * n;

                for (int chPos = 0; chPos < halfWidthHeight; chPos++)
                {
                    if (chosenPositions[chPos + 1] > (chosenPositions[chPos] + 1))
                    {
                        chosenPositions[chPos]++;
                        for (int ind = 0; ind < chPos; ind++)
                            chosenPositions[ind] = ind;

                        break;
                    }

                }

                regionsCalc++;

                // if (regionsCalc % 10000 == 0)
                // Console.WriteLine("{0}: sum = {1}", stopWatch.Elapsed, sum.ToString("N0"));

            }
            Task.WaitAll(tasks);
            Console.WriteLine("{0} -- Finished all tasks", DateTime.Now);
            foreach (Task<long> t in tasks)
                sum += (t.Result * t.Result);

         
            Console.WriteLine("Regions: {0}", regionsCalc);
            Console.WriteLine("Tilings: {0}", sum.ToString("N0"));
            Console.WriteLine("Finished: {0}", DateTime.Now);
            return sum;
        }

        #endregion

        #region Parallel Flip Methods for Boxes


        /// <summary>
        /// Finds the flip connected component of a box. Uses some parallel computing tools.
        /// </summary>
        /// <param name="boxDim">Dimensions of the box</param>
        /// <returns>>A LargeConnectedComponentsInfo object with information on the connected components.</returns>
        public static LargeConnectedComponentsInfo ParallelFindFlipCC(BoxDimensions boxDim)
        {
            return ParallelFindFlipCC(boxDim.Width, boxDim.Height, boxDim.Depth);
        }

        /// <summary>
        /// Finds the flip connected component of a box. Uses some parallel computing tools.
        /// </summary>
        /// <param name="width">The width of the box</param>
        /// <param name="height">The height of the box </param>
        /// <param name="depth">The depth of the box</param>
        /// <returns>A LargeConnectedComponentsInfo object with information on the connected components.</returns>
        public static LargeConnectedComponentsInfo ParallelFindFlipCC(int width, int height, int depth)
        {
            #region Initializing main variables


            if (depth % 2 != 0)
            {
                if (width % 2 == 0)
                    return ParallelFindFlipCC(depth, height, width);
                else if (height % 2 == 0)
                    return ParallelFindFlipCC(width, depth, height);
                else
                    return null;
            }
            Console.WriteLine("{3} - Started to find FCC of a {0}x{1}x{2} box", width, height, depth, DateTime.Now);
            int halfDepth = depth >> 1;
            int WidthHeigth = width * height;
            int halfWidthHeight = (halfDepth % 2 == 0 ? (WidthHeigth >> 1) : (WidthHeigth + 1) >> 1);



            //long sum = 0;

            int[] chosenPositions = new int[halfWidthHeight + 1]; //The last one is an auxiliary construction
            for (int i = 0; i < halfWidthHeight; i++)
            {
                chosenPositions[i] = i;
            }

            chosenPositions[halfWidthHeight] = WidthHeigth + 10;

            int regionsCalc = 0;
            int totalRegions = (int)MathConvenience.nChoosek(WidthHeigth, halfWidthHeight);
            Console.WriteLine("Number of regions to calc: {0}", totalRegions.ToString("N0"));

            Task<Dictionary<string, UnionFindNode<string>>>[] tasks = new Task<Dictionary<string, UnionFindNode<string>>>[totalRegions];

            Dictionary<string, int> regionToTaskIndex = new Dictionary<string, int>(totalRegions);
            #endregion

            #region Setting base region and tasks

            while (chosenPositions[halfWidthHeight - 1] < WidthHeigth)
            {
                Region baseRegion = new Region(new BoxDimensions(width, height, depth));
                int i = 0;
                for (int chPos = 0; chPos < halfWidthHeight; chPos++)
                {

                    while (i < chosenPositions[chPos])
                    {
                        baseRegion[i] = (baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);
                        i++;
                    }


                    baseRegion[i] = (!baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);
                    i++;

                }

                for (; i < WidthHeigth; i++)
                    baseRegion[i] = (baseRegion.IsWhite(i) ? DimerDirection.Empty : DimerDirection.Inaccessible);

                regionToTaskIndex.Add(baseRegion.ToString(), regionsCalc);

                tasks[regionsCalc] = Task<Dictionary<string, UnionFindNode<string>>>.Factory.StartNew((region) =>
                {
                    Region r = region as Region;

                    return r.FindFlipConnectedComponents();

                }, baseRegion);


                for (int chPos = 0; chPos < halfWidthHeight; chPos++)
                {
                    if (chosenPositions[chPos + 1] > (chosenPositions[chPos] + 1))
                    {
                        chosenPositions[chPos]++;
                        for (int ind = 0; ind < chPos; ind++)
                            chosenPositions[ind] = ind;

                        break;
                    }

                }


                regionsCalc++;

            }
            Task.WaitAll(tasks);
            Console.WriteLine("{0} -- Finished all tasks", DateTime.Now);
            //string fileName = @"C:\Users\phmilet\Documents\Doutorado\TilingsPrograms\FCC_44\fccDics.txt";

            #endregion

            #region Calculating all parents in half boxes

            Dictionary<StringPair, UnionFindNode<StringPair>> fullBoxParents = new Dictionary<StringPair, UnionFindNode<StringPair>>(new StringPairEqualityComparer());

            foreach (Task<Dictionary<string, UnionFindNode<string>>> t in tasks)
            {
                //Dictionary<string, List<string>> parentsDic = new Dictionary<string, List<string>>();
                Dictionary<string, long> parentsDic = new Dictionary<string, long>();
                Dictionary<string, UnionFindNode<string>> ufdic = t.Result;


                foreach (string til in ufdic.Keys)
                {
                    UnionFindNode<string> tNode = (ufdic[til].Find()) as UnionFindNode<string>;
                    string parent = tNode.Value;

                    if (parentsDic.ContainsKey(parent))
                    {
                        //parentsDic[parent].Add(til);
                        parentsDic[parent]++;

                    }
                    else
                    {
                        //parentsDic.Add(parent, new List<string>());
                        //parentsDic[parent].Add(til);

                        parentsDic.Add(parent, 1);

                    }


                }


                foreach (string par in parentsDic.Keys)
                {
                    foreach (string secondPar in parentsDic.Keys)
                    {
                        StringPair sp = new StringPair(par, secondPar, parentsDic[par] * parentsDic[secondPar]);
                        fullBoxParents.Add(sp, new UnionFindNode<StringPair>(sp));
                    }
                }



            }

            Console.WriteLine("{0} - Finished parent and StringPair calculations", DateTime.Now);

            //TEST CODE:
            /*
            long summand = 0;
            foreach (StringPair sp in fullBoxParents.Keys)
            {
                summand += sp.NumberOfElements;
            }

            Console.WriteLine("In principle, total number of elements = {0}", summand);

             * */
            #endregion

            #region Connecting parents in different boxes

            foreach (StringPair sp in fullBoxParents.Keys)
            {
                Region firstTiling = Region.FromString(sp.Item1);
                Region secondTiling = Region.FromString(sp.Item2);

                for (int pos = 0; pos < firstTiling.Length; pos++)
                {
                    if (firstTiling[pos] == DimerDirection.Inaccessible)
                    {
                        int i, j, k;
                        firstTiling.IndexVariables(pos, out i, out j, out k);

                        //If the tile to the right of the current is inaccessible, then a flip is possible
                        if (i + 1 < firstTiling.Width && firstTiling[i + 1, j, k] == DimerDirection.Inaccessible)
                        {
                            firstTiling[i, j, k] = DimerDirection.Right;
                            firstTiling[i + 1, j, k] = DimerDirection.Left;

                            secondTiling[i, j, k] = DimerDirection.Right;
                            secondTiling[i + 1, j, k] = DimerDirection.Left;

                            string regString = Region.EmptyString(firstTiling.ToString());
                            Dictionary<string, UnionFindNode<string>> ufdic = tasks[regionToTaskIndex[regString]].Result;


                            string firstParent = ufdic[firstTiling.ToString()].Find().Value;
                            string secondParent = ufdic[secondTiling.ToString()].Find().Value;
                            StringPair newSP = new StringPair(firstParent, secondParent);

                            UnionFindNode<string>.Union(fullBoxParents[sp], fullBoxParents[newSP]);

                            firstTiling[i, j, k] = DimerDirection.Inaccessible;
                            firstTiling[i + 1, j, k] = DimerDirection.Inaccessible;

                            secondTiling[i, j, k] = DimerDirection.Inaccessible;
                            secondTiling[i + 1, j, k] = DimerDirection.Inaccessible;

                        }

                        //If the tile downward from the current one is inaccessible, a flip is possible
                        if (j + 1 < firstTiling.Height && firstTiling[i, j + 1, k] == DimerDirection.Inaccessible)
                        {
                            firstTiling[i, j, k] = DimerDirection.Down;
                            firstTiling[i, j + 1, k] = DimerDirection.Up;

                            secondTiling[i, j, k] = DimerDirection.Down;
                            secondTiling[i, j + 1, k] = DimerDirection.Up;

                            string regString = Region.EmptyString(firstTiling.ToString());
                            Dictionary<string, UnionFindNode<string>> ufdic = tasks[regionToTaskIndex[regString]].Result;

                            string firstParent = ufdic[firstTiling.ToString()].Find().Value;
                            string secondParent = ufdic[secondTiling.ToString()].Find().Value;
                            StringPair newSP = new StringPair(firstParent, secondParent);

                            UnionFindNode<string>.Union(fullBoxParents[sp], fullBoxParents[newSP]);

                            firstTiling[i, j, k] = DimerDirection.Inaccessible;
                            firstTiling[i, j + 1, k] = DimerDirection.Inaccessible;

                            secondTiling[i, j, k] = DimerDirection.Inaccessible;
                            secondTiling[i, j + 1, k] = DimerDirection.Inaccessible;
                        }
                    }
                }
            }

            Console.WriteLine("{0} - Finished CC calculations - now writing to file", DateTime.Now);

            #endregion

            #region Processing parents of the full box

            //string fileName = @"D:\Doutorado\TilingsFiles\perfectTests.txt";

            //HashSet<StringPair> knownParents = new HashSet<StringPair>(new StringPairEqualityComparer());
            Dictionary<StringPair, long> knownParents = new Dictionary<StringPair, long>(new StringPairEqualityComparer());

            //Now we need to extract the parents from this huge dictionary
            //using (StreamWriter sw = new StreamWriter(fileName))
            //{
            //Dictionary<string, List<string>> parentsDic = new Dictionary<string, List<string>>();

            foreach (UnionFindNode<StringPair> uf in fullBoxParents.Values)
            {
                StringPair sp = uf.Find().Value;

                if (knownParents.ContainsKey(sp))
                {
                    knownParents[sp] += uf.Value.NumberOfElements;
                }
                else
                {
                    knownParents.Add(sp, uf.Value.NumberOfElements);
                }

            }

            //    Console.WriteLine("FullBox Count: {0}, HashCount: {1}", fullBoxParents.Count, knownParents.Count);

            //    sw.WriteLine("Number of CCs: {0}", knownParents.Count);
            //    foreach (StringPair sp in knownParents.Keys)
            //    {
            //        sw.WriteLine("New CC parent: {0} + {1}", sp.Item1, sp.Item2);
            //        sw.WriteLine("Number of tilings: {0}", knownParents[sp].ToString("N0"));
            //        sw.WriteLine();
            //    }

            //    sw.Flush();

            //}



            //Console.WriteLine("Regions: {0}", regionsCalc);
            //Console.WriteLine("Tilings: {0}", sum.ToString("N0"));
            Console.WriteLine("Finished: {0}", DateTime.Now);

            #endregion

            #region Finishing up and returning

            Dictionary<string, Dictionary<string, UnionFindNode<string>>> regionToCCDic = new Dictionary<string, Dictionary<string, UnionFindNode<string>>>(totalRegions);

            foreach (string reg in regionToTaskIndex.Keys)
            {
                regionToCCDic.Add(reg, tasks[regionToTaskIndex[reg]].Result);
            }

            LargeConnectedComponentsInfo componentsInfo = new LargeConnectedComponentsInfo()
            {
                RegionToUFDic = regionToCCDic,
                KnownParents = knownParents,
                FullBoxParents = fullBoxParents
            };

            return componentsInfo;

            #endregion

        }

        #endregion
    }



    #region Math Convenience class

    static class MathConvenience
    {
        public static long nChoosek(int n, int k)
        {
            long prod = 1;
            long denom = 1;
            while (k > 0)
            {
                prod *= n;
                denom *= k;

                n--;
                k--;
            }

            return (prod / denom);
        }
    }

    #endregion



}
